# -*- coding: utf_8 -*-
"""
ize 测试功能示例; everything in this package have no compability
"""
from .botanical_clock import botanical_clock  # noqa
from .dancing_example import dancing_example  # noqa
from .end_callback_example import end_test  # noqa
from .pole_jumping_test import pole_jumping_test  # noqa
from .iztools_example import default_test  # noqa
